@@ -49,10 +49,19 @@ module What4.Protocol.SMTLib2.Syntax
   , assertNamed
   , getUnsatAssumptions
   , getUnsatCore
+  , getAbduct
+  , getAbductNext
+    -- * SyGuS
+  , synthFun
+  , declareVar
+  , constraint
+  , checkSynth
     -- * Logic
   , Logic(..)
   , qf_bv
   , allSupported
+  , allLogic
+  , hornLogic
     -- * Sort
   , Sort(..)
   , boolSort
@@ -184,6 +193,15 @@ qf_bv = Logic "QF_BV"
 -- | Set the logic to all supported logics.
 allSupported :: Logic
 allSupported = Logic "ALL_SUPPORTED"
+{-# DEPRECATED allSupported "Use allLogic instead" #-}
+
+-- | Set the logic to all supported logics.
+allLogic :: Logic
+allLogic = Logic "ALL"
+
+-- | Use the Horn logic
+hornLogic :: Logic
+hornLogic = Logic "HORN"
 
 ------------------------------------------------------------------------
 -- Symbol
@@ -290,7 +308,7 @@ or [] = true
 or [x] = x
 or l = term_app "or" l
 
--- | Disjunction of all terms
+-- | Xor of all terms
 xor :: [Term] -> Term
 xor l@(_:_:_) = term_app "xor" l
 xor _ = error "xor expects two or more arguments."
@@ -299,7 +317,7 @@ xor _ = error "xor expects two or more arguments."
 eq :: [Term] -> Term
 eq = chain_app "="
 
--- | Construct a chainable term with the givne relation
+-- | Construct a chainable term with the given relation
 --
 -- @pairwise_app p [x1, x2, ..., xn]@ is equivalent to
 -- \forall_{i,j} p x_i x_j@.
@@ -480,7 +498,7 @@ isInt = un_app "is_int"
 -- value type `t2` that always returns `c`.
 --
 -- This uses the non-standard SMTLIB2 syntax
--- @((as const (Array t1 t2)) c)@ which is supported by CVC4 and Z3
+-- @((as const (Array t1 t2)) c)@ which is supported by CVC4, CVC5, and Z3
 -- (and perhaps others).
 arrayConst :: Sort -> Sort -> Term -> Term
 arrayConst itp rtp c =
@@ -820,6 +838,37 @@ getUnsatAssumptions = Cmd "(get-unsat-assumptions)"
 getUnsatCore :: Command
 getUnsatCore = Cmd "(get-unsat-core)"
 
+-- | Get an abduct that entails the formula, and bind it to the name
+getAbduct :: Text -> Term -> Command
+getAbduct nm p = Cmd $ "(get-abduct " <> Builder.fromText nm <> " " <> renderTerm p <> ")"
+
+-- | Get the next command, called after a get-abduct command
+getAbductNext :: Command
+getAbductNext = Cmd "(get-abduct-next)"
+
+-- | Declare a SyGuS function to synthesize with the given name, arguments, and
+-- return type.
+synthFun :: Text -> [(Text, Sort)] -> Sort -> Command
+synthFun f args ret_tp = Cmd $ app "synth-fun"
+  [ Builder.fromText f
+  , builder_list $ map (\(var, tp) -> app (Builder.fromText var) [unSort tp]) args
+  , unSort ret_tp
+  ]
+
+-- | Declare a SyGuS variable with the given name and type.
+declareVar :: Text -> Sort -> Command
+declareVar v tp = Cmd $ app "declare-var" [Builder.fromText v, unSort tp]
+
+-- | Add the SyGuS constraint to the current synthesis problem.
+constraint :: Term -> Command
+constraint p = Cmd $ app "constraint" [renderTerm p]
+
+-- | Ask the SyGuS solver to find a solution for the synthesis problem
+-- corresponding to the current functions-to-synthesize, variables and
+-- constraints.
+checkSynth :: Command
+checkSynth = Cmd "(check-synth)\n"
+
 -- | Get the values associated with the terms from the last call to @check-sat@.
 getValue :: [Term] -> Command
 getValue values = Cmd $ app "get-value" [builder_list (renderTerm <$> values)]
